from __future__ import annotations

import datetime
import math
import time
import typing as t

DEFAULT_TIMER_PRECISION: t.Final = 2


class Timer:
    """
    Initializes a Timer instance with optional precision.

    Args:
        precision (int): The number of decimal places to use
            for displaying fractional seconds. Defaults to 2.
    """

    def __init__(self, precision: int = DEFAULT_TIMER_PRECISION) -> None:
        self.precision = precision
        self.start: float | None = None
        self.end: float | None = None

    def __enter__(self) -> Timer:
        self.start = time.time()
        return self

    def __exit__(self, *args: t.Any) -> None:
        self.end = time.time()

    def __str__(self) -> str:
        return self.verbose()

    def __repr__(self) -> str:
        return f"<Timer [start={self.start}, end={self.end}]>"

    def verbose(self) -> str:
        """
        Returns a formatted string representing the elapsed time with a precision
        specified by the Timer instance.

        The elapsed time is formatted as a string in the format of 'H:MM:SS.F',
        where 'H:MM:SS' is the hours, minutes, and seconds, and 'F' is the
        fractional seconds with a number of decimal places equal to the precision.

        Returns:
            str: The formatted elapsed time as a string.
        """
        fraction_seconds, whole_seconds = math.modf(self.elapsed)
        rounded_fraction = round(fraction_seconds, self.precision)
        if rounded_fraction >= 1:
            whole_seconds += 1
            formatted_fraction = "0" * self.precision
        elif fraction_seconds == 0:
            formatted_fraction = "0" * self.precision
        else:
            fraction = int(rounded_fraction * 10**self.precision)
            formatted_fraction = str(fraction).zfill(self.precision)
        return f"{datetime.timedelta(seconds=whole_seconds)}.{formatted_fraction}"

    @property
    def elapsed(self) -> float:
        """
        Calculates the elapsed time in seconds.

        This property computes the difference between the end time and the
        start time of the Timer. If the Timer has not been started, it returns
        0.0. If the Timer is running (i.e., the end time is not set), it uses
        the current time as the end time.

        Returns:
            float: The elapsed time in seconds.
        """
        if not self.start:
            return 0.0
        return (self.end or time.time()) - self.start



import typing as t
from collections.abc import Sequence
from functools import partial

T = t.TypeVar("T")
K = t.TypeVar("K")

DEFAULT_ID_ATTR_NAME: t.Final[str] = "id"

#: List of values that are considered empty
EMPTY_VALUES: t.Final[list[t.Any]] = ["", None, [], {}]


def chunker(seq: Sequence[T], size: int) -> list[Sequence[T]]:
    """Splits provided iterable into list of chunks of given size

    Args:
        seq: iterable to split
        size: chunk size

    Returns:
        list of chunks
    """
    return [seq[pos : pos + size] for pos in range(0, len(seq), size)]


@t.overload
def idfy(
    obj: dict[t.Any, t.Any], id_field_name: str = DEFAULT_ID_ATTR_NAME
) -> dict[t.Any, dict[t.Any, t.Any]]: ...


@t.overload
def idfy(obj: list[T], id_field_name: str = DEFAULT_ID_ATTR_NAME) -> dict[t.Any, T]: ...


@t.overload
def idfy(obj: set[T], id_field_name: str = DEFAULT_ID_ATTR_NAME) -> dict[t.Any, T]: ...


@t.overload
def idfy(obj: tuple[T, ...], id_field_name: str = DEFAULT_ID_ATTR_NAME) -> dict[t.Any, T]: ...


@t.overload
def idfy(obj: T, id_field_name: str = DEFAULT_ID_ATTR_NAME) -> dict[t.Any, T]: ...


def idfy(
    obj: T | list[T] | set[T] | tuple[T, ...] | dict[t.Any, t.Any],
    id_field_name: str = DEFAULT_ID_ATTR_NAME,
) -> dict[t.Any, T] | dict[t.Any, dict[t.Any, t.Any]]:
    """Converts given object into dict with ``id_field_name`` values as a key
    and actual object as a value.

    If given object is a dict this function uses to get value of
    ``id_field_name`` key. If any other object - looks for ``id_field_name``
    attribute.

    Raises :py:exc:`ValueError` if there's no appropriate id value found.

    Applied recursively if :py:class:`Iterable` is given as an input.

    Args:
        obj (T): object to convert to dictionary
        id_field_name (str): name of the field/attribute to use to get, defaults
            to `"id"`

    Returns:
        dict with ``id_field_name`` values as a key and actual object as a value
    """
    if isinstance(obj, dict):
        try:
            return {obj[id_field_name]: obj}
        except KeyError as err:
            raise ValueError(f"Can't get '{id_field_name}' key from {obj} dict") from err
    if isinstance(obj, (list, set, tuple, t.Generator)):
        return {k: v for d in obj for k, v in idfy(d, id_field_name).items()}
    if hasattr(obj, id_field_name):
        return {getattr(obj, id_field_name): obj}
    raise ValueError(f"Can't get {id_field_name} attribute from {obj}")


@t.overload
def remove_empty_members(
    obj: dict[t.Any, t.Any], empty: list[t.Any] | None = None
) -> dict[t.Any, t.Any] | None: ...


@t.overload
def remove_empty_members(
    obj: list[t.Any], empty: list[t.Any] | None = None
) -> list[t.Any] | None: ...


@t.overload
def remove_empty_members(obj: T, empty: list[t.Any] | None = None) -> T | None: ...


def remove_empty_members(obj: T, empty: list[t.Any] | None = None) -> T | None:
    """Removes empty members from given object.

    Recursively goes through the given object and removes its members that are
    considered empty. If given object is a dict, removes keys that have empty
    values. If given object is a list removes empty items from it.

    Args:
        obj: object to remove empty members from
        empty: list of values that are considered empty. If not given, defaults to
            :const:`EMPTY_VALUES`

    Returns:
        object with empty members removed, or None if the object itself is empty
    """
    empty = empty or EMPTY_VALUES
    cleaners: dict[type, t.Callable[[t.Any, list[t.Any]], t.Any]]
    cleaners = {
        dict: _remove_empty_members_from_dict,
        list: _remove_empty_members_from_list,
    }
    if (cleaner := cleaners.get(type(obj))) is not None:
        obj = cleaner(obj, empty)
    if obj in empty:
        return None
    return obj


def _remove_empty_members_from_dict(
    obj: dict[t.Any, t.Any], empty: list[t.Any]
) -> dict[t.Any, t.Any]:
    for key in list(obj.keys()):
        obj[key] = remove_empty_members(obj[key], empty)
        if obj[key] is None:
            del obj[key]
    return obj


def _remove_empty_members_from_list(obj: list[t.Any], empty: list[t.Any]) -> list[t.Any]:
    return list(
        filter(lambda x: x not in empty, map(partial(remove_empty_members, empty=empty), obj))
    )



import logging
from functools import cached_property


class LoggerMixin:  # pylint: disable=too-few-public-methods
    """Simple mixin for logging.

    Adds a ``logger`` property to the class and sets the logger name
    to the module and class name of the class:

        >>> class MyClass(LoggerMixin):
        ...     pass
        ...
        >>> my_class = MyClass()
        >>> my_class.logger.name
        'mymodule.MyClass'
    """

    @cached_property
    def logger(self) -> logging.Logger:
        """Logger instance for this class.

        The logger name is determined by the module and class name of this class.
        For example, if the class is named `MyClass` and it's in the module
        `mymodule`, the logger name will be `mymodule.MyClass`.

        Returns:
            Logger: Logger instance for this class.
        """
        name = ".".join([self.__class__.__module__, self.__class__.__name__])
        return logging.getLogger(name)


def set_level_for_logger(
    logger_names: str | list[str] | tuple[str, ...], level: int = logging.WARNING
) -> None:
    """Change logging level for one or more loggers.

    Args:
        logger_names (str | list[str] | tuple[str, ...]): Logger name(s) to change level for.
        level (int, optional): Logging level to set. Defaults to ``logging.WARNING``.

    Returns:
        None
    """
    if isinstance(logger_names, str):
        logging.getLogger(logger_names).setLevel(level)
    elif isinstance(logger_names, (tuple, list)):
        set_level_for_logger(logger_names, level)
    raise TypeError(
        f"Expected logger_names to be str, tuple or list, got {type(logger_names)} instead"
    )


def get_simple_logger(
    name: str = "root",
    console: bool = True,
    log_file: str | None = None,
    log_level: str | int = logging.DEBUG,
) -> logging.Logger:
    """Creates simple instance of logger.

    Useful for simple scripts, where precise configuration is not needed. Output
    example:

        >>> import logging
        >>> logger = get_simple_logger()
        >>> logger.debug("Hello, world!")
        2023-05-01T12:34:56 root DEBUG Hello, world!

    Args:
        name (str | None, optional): Name of logger. Defaults to "root".
        console (bool, optional): Add console handler. Defaults to True.
        log_level (str | int, optional): Log level. Defaults to ``logging.DEBUG``.
        log_file (str | None, optional): File name to write log to. Defaults to None.

    Returns:
        logging.Logger
    """
    log_formatter = logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s")
    log_formatter.datefmt = "%Y-%m-%dT%H:%M:%S"
    logger = logging.getLogger()
    logger.name = name
    logger.setLevel(log_level)
    if log_file:
        file_handler = logging.FileHandler(log_file, mode="a")
        file_handler.setFormatter(log_formatter)
        logger.addHandler(file_handler)

    if console:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level)
        console_handler.setFormatter(log_formatter)
        logger.addHandler(console_handler)
    return logger



import datetime


def days_after_now(
    days_num: int = 1, tz: datetime.tzinfo | None = datetime.timezone.utc
) -> datetime.datetime:
    """Returns a datetime object representing the date that is ``days_num`` days after now.

    Args:
        days_num (int): The number of days after now. Defaults to 1.
        tz (datetime.tzinfo, optional): The timezone to use. Defaults to
            :attr:`datetime.timezone.utc`.

    Returns:
        The datetime that is ``days_num`` days after now.
    """
    return datetime.datetime.now(tz=tz) + datetime.timedelta(days=days_num)


def tomorrow(tz: datetime.tzinfo | None = datetime.timezone.utc) -> datetime.datetime:
    """Returns a datetime object representing the date that is 1 day after now.

    Args:
        tz (datetime.tzinfo, optional): The timezone to use. Defaults to
            :attr:`datetime.timezone.utc`.

    Returns:
        The datetime that is 1 day after now.
    """
    return days_after_now(tz=tz)


def days_before_now(
    days_num: int = 1, tz: datetime.tzinfo | None = datetime.timezone.utc
) -> datetime.datetime:
    """Returns a datetime object representing the date that is ``days_num`` days before now.

    Args:
        days_num (int): The number of days before now. Defaults to 1.
        tz (datetime.tzinfo, optional): The timezone to use. Defaults to
            :attr:`datetime.timezone.utc`.

    Returns:
        The datetime that is ``days_num`` days before now.
    """
    return days_after_now(-days_num, tz=tz)


def yesterday(tz: datetime.tzinfo | None = datetime.timezone.utc) -> datetime.datetime:
    """Returns a datetime object representing the date that is 1 day before now.

    Args:
        tz (datetime.tzinfo, optional): The timezone to use. Defaults to
            :attr:`datetime.timezone.utc`.

    Returns:
        The datetime that is 1 day before now.
    """
    return days_before_now(tz=tz)


def yesterday_start(tz: datetime.tzinfo | None = datetime.timezone.utc) -> datetime.datetime:
    """Returns a datetime object representing the start of yesterday.

    Args:
        tz (datetime.tzinfo, optional): The timezone to use. Defaults to
            :attr:`datetime.timezone.utc`.

    Returns:
        The datetime that is the start of yesterday.
    """
    return datetime.datetime.combine(yesterday(tz=tz), datetime.datetime.min.time()).replace(
        tzinfo=tz
    )


def yesterday_end(tz: datetime.tzinfo | None = datetime.timezone.utc) -> datetime.datetime:
    """
    Returns a datetime object representing the end of yesterday.

    Args:
        tz (datetime.tzinfo, optional): The timezone to use. Defaults to
            :attr:`datetime.timezone.utc`.

    Returns:
        The datetime that is the end of yesterday.
    """
    return datetime.datetime.combine(yesterday(tz=tz), datetime.datetime.max.time()).replace(
        tzinfo=tz
    )


def today_eod(tz: datetime.tzinfo = datetime.timezone.utc) -> datetime.datetime:
    """
    Returns a datetime object representing the end of the current day.

    Args:
        tz (datetime.tzinfo, optional): The timezone to use. Defaults to
            :attr:`datetime.timezone.utc`.

    Returns:
        The datetime that is the end of the current day.
    """
    today = datetime.datetime.now().date()
    return datetime.datetime.combine(today, datetime.datetime.max.time()).replace(tzinfo=tz)



import typing as t

SIZE_UNIT: t.Final = "B"
DEFAULT_PRECISION: t.Final = 2


def size(size_bytes: int | float, precision: int = DEFAULT_PRECISION) -> str:
    """
    Converts the given size in bytes into a human-friendly string representation.

    Args:
        size_bytes (int or float): The size in bytes to convert.
        precision (int, optional): The number of decimal places to round the
            result to. Defaults to 2.

    Returns:
        str: A human-friendly string representation of the given size,
            e.g. 4.00 B, 1.00 KiB, 1.23 MiB, etc.
    """
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(size_bytes) < 1024.0:
            return f"{size_bytes:.{precision}f} {unit}{SIZE_UNIT}"
        size_bytes /= 1024.0
    return f"{size_bytes:.{precision}f} Yi{SIZE_UNIT}"
